#  FileQueue.pm
#    - providing a simple object for safe (lockable) file queue arbritary data
#
#  Copyright (C) 2004 Doolin Technologies
#
#  $Id: FileQueue.pm,v 1.1 2004/12/17 17:39:03 jamesc Exp $
#

package FileQueue;

$VERSION = '0.06';

use strict;
use bytes;

use Queue::Dir;
use Storable;
use Storable qw(store_fd fd_retrieve);

sub new($%) {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my %args = @_;
    my $self = {};
    my $errmark = "FileQueue::new error: ";

    #$self->{debug} = 1;
    #$self->{verbose} = 1;
    $self->{debug} = ($args{debug})?$args{debug}:0;
    $self->{verbose} = ($args{verbose})?$args{verbose}:0;

    if ($args{dir}) {
	$self->{dir} = $args{dir};
	mkdir $self->{dir} if (!-d $self->{dir});
        #why would a user be able to access (r&w via group permissions) a dir fine but a service running as that user not? :'( 
	if (!-d $self->{dir} || !-r $self->{dir} || !-w $self->{dir}) {
	    my $err = $errmark . "check permissions. ".
		"can't access or create queue directory: -" . $self->{dir} . "-";
	    $err .= "\ngroups: " . `groups`;
	    $err .= "\nwhoami: " . `whoami`;
	    print STDERR "$err\n";
	    return { error => 1,
		     errors => [ $err ]
		     };
	}
    } else {
	my $err = $errmark . "dir arg is mandatory";
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err ]
		 };
    }

    # TODO setting queue id like this could cause confusion with 
    # multiple Qs and same queue id ?
    # I set id to dir before but ... 
    # was okay in new and not when I wanted to store something in q
    # queue id
    $self->{id} = ($args{id})?$args{id}:"defaultqname".time();
    # queue lock dir
    $self->{lock} = ($args{lock})?$args{lock}:"lock";

    $self->{q} = new Queue::Dir  
	id => $self->{id},
	paths => [ $self->{dir} ],
	promiscuous => 1,
	#sort => 'sortsub',
	#filter => sub { ... },
	lockdir => $self->{lock},
	#lockmax => 300, # default 300 sec anyway
	;

    if (!defined $self->{q}) {
	my $err = $errmark . "dir arg is mandatory";
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err ]
		 };
    }

    $self->{q}->unlock;
    $self->{lastid} = 0;

    # TODO can we leave q instantiated ?
    if (0) {
	$self->{q} = undef;
    }

    bless $self, $class;
}

# TODO: review locks
#  locks are specific to $qid
#  careful visit locks stuff, next ?doesn't?
# visit locks stuff but does not return qid.
# unlock with no params unlocks current thing
# But if gone one to another thing cannot remove lock
# (no way to call unlock qith a qid => multiple locks not possible :( )

# provided methods:

# my $res = $smsfilequeue->enqueue( $req );
# $smsfilequeue->reset();  # so readnext will start with highest priority
# while ($smsfilequeue->notempty()) { # blocks and waits for lock ?
#     my ($id, $req) = $smsfilequeue->readnext();
#	 $smsfilequeue->dequeue($id);
#	 $smsfilequeue->update($id,{ yokeybob => "thingybob" });

# in practice we will process Q entries one by one
# we will never in middle of q wish to reset it I think
# if ever need to reset in practice better to use visit
#   actually I think visit, if half way through a queue would not find qid already passed
# another side-effect to note is Queue::Dir refreshes queue from disk when it gets to end of Q
sub reset {
    my ($self) = @_;

    my $errmark = "FileQueue::reset ";
    print $errmark if ($self->{debug});

    # unlock last thing # TODO what if last thing we did was a remove?
    $self->{q}->unlock();

    my $qid = $self->{q}->next();
    while (defined($qid) && $qid) {
	#$fh = $self->{q}->visit;
	#my $ref = fd_retrieve $fh;
	$qid = $self->{q}->next();
    }

    $self->{lastid} = 0;

}

sub whong {
    my ($self) = @_;

    my $errmark = "FileQueue::whong ";
    print $errmark if ($self->{debug});

    $self->{q}->unlock();
    $self->{q} = undef;

    $self->{q} = new Queue::Dir  
	id => $self->{id},
	paths => [ $self->{dir} ],
	promiscuous => 1,
	lockdir => $self->{lock},
	;

    if (!defined($self->{q})) {
	my $err = $errmark . "fail to open Q: $!";
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err ]
		 };
    }

    $self->{lastid} = 0;

}

#  side effect readnext leaves last thing read locked
# if I read - I should really leave it locked and unlock it myself
# when we're processing q we wish to 
#   read thing, 
#   process and rm or 
#   fail to process and perhaps update
#   then unlock    

# read next or read entry matching id
# visit locks the qid
# remember to unlock 
sub readnext {
    my ($self) = shift;
    my $id     = shift; # may be undef

    my $errmark = "FileQueue::readnext ";
    print STDERR $errmark if ($self->{debug});

    ### careful

    ### 1. next wraps queue with a refresh

    ### 2. a visit calls next itself if qid not defined and current not defined!
    ### so we must be careful to not accidentally call next twice 
    ###   e.g. at enf of processing a q

    ### 3. a visit call SKIPS over locked queue entries and retruns next unlocked found
    ###    THUS last qid we got with next may not match entry returned
    ###    TODO: test can we safely unlock a thing after visit skips over locked things
    # #           (internally current qid doesn't match ??? no. couldn't be)
 
    my ($qid,$fh);
    $qid = $self->{q}->next($id);
    if (defined($qid) && $qid 
	#&& $qid != $self->{lastid}
	) {
	$fh = $self->{q}->visit("r",$id);

	# TODO: test $fh, is it a valid file descriptor?

	if (defined($fh) && $fh) {
	    my $ref = fd_retrieve $fh;
	    #$self->{lastid} = $qid;
	    return { id => $qid,
		     ref => $ref };
	} else {
	    print STDERR $errmark."fh valid file desc " .
		defined($fh)?$fh:"" . "\n"
		if ($self->{debug});
	    #return { error => "problem with file handle. cannot visit file" } ;
	    # do not return undef - other things in q ?
            # $fh can be 0 ? yes. why?
	    # TODO: nasty permissions problem
	}

    } 

    return undef;
}

sub unlock {
    my $self = shift;
    my $id = shift || undef;
    # unlock last thing # TODO what if last thing we did was a remove?
    $self->{q}->unlock($id);
}

sub enqueue {
    my ( $self, $reftothing) = @_;
    #my %args = @_;

    my $errmark = "FileQueue::enqueue ";
    print STDERR $errmark . Dumper($reftothing) if ($self->{debug});
    print STDERR $errmark . "id:" . $self->{id} . "dir:" . $self->{dir} .
	"lock:" . $self->{lock} . "\n" if ($self->{debug});

    if (0) {
	$self->{q} = new Queue::Dir  
	    id => $self->{id},
	    paths => [ $self->{dir} ],
	    promiscuous => 1,
	    lockdir => $self->{lock},
	    ;

	if (!defined($self->{q})) {
	    my $err = $errmark . "fail to open Q: $!";
	    print STDERR "$err\n";
	    return { error => 1,
		     errors => [ $err ]
		     };
	}
    }

    if (0) {
	my ($fh, $qid);
	while (!defined($qid) || !defined($fh)) {
	    ($fh, $qid) = $self->{q}->store();
	    sleep 1;
	}
    }

    my ($fh, $qid) = $self->{q}->store();

    #my ($fh, $qid) = $self->{q}->store();
    if (!defined($qid) || !defined($fh)) {
	my $err = $errmark . "fail to open storable filehandle";
	$err .= ", qid is " . (defined($qid))?$qid:"not defined";
	$err .= ", fh is " . (defined($fh))?$fh:"not defined";
	#$err .= "carp?" . eval {} ;
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err ]
		 };
    }

    #(ref $fh eq 'IO::File', "Proper fh");
    #print $fh "and $qid\n";

    store_fd $reftothing, $fh;

    print "id $self->{id}, qid $qid, name:" . $self->{q}->name($qid) . "\n"
	if ($self->{debug});

    $fh->close;
    $self->{q}->unlock;
    $qid = undef; 
    $fh = undef;

    if (0) {
	$self->{q} = undef; # delete things
    }

    #return { error => 0, id => $qid };
    return { error => 0 };
}

# remove item from Q
sub dequeue {
    my ( $self, $id) = @_;

    my $errmark = "FileQueue::dequeue ";
    print STDERR $errmark . $id if ($self->{debug});

    my $status = $self->{q}->done($id);
    #print STDERR "item rm from queue $status\n";
    $self->{q}->unlock($id);

    return $status;
}

# from current item to end
sub show_queue{
    my ($self, $print) = @_;

    my $errmark = "FileQueue::show_queue ";
    print STDERR $errmark if ($self->{debug});

    my ($qid,$fh);

    use Data::Dumper;

    my $count=0;
    $qid = $self->{q}->next();
    while (defined($qid)) {
	$fh = $self->{q}->visit;
	if (defined($fh) && $fh) {
	    my $ref = fd_retrieve $fh;

	    #print Dumper($ref);
	    print STDERR "qid $qid, name:" . $self->{q}->name($qid) . 
		", text: ".$ref->{text}."\n" if ($print);

	    $self->{q}->unlock; # visit locks thing
	} else {
	    print STDERR $errmark."fh valid file desc?\n";
	}
	# TODO: test $fh, is it a valid file descriptor?

	$qid = $self->{q}->next();
	$count++;
    }

    return $count;
}

sub update {
    my ($self) = shift;
    my $id     = shift; # may be undef BUT that would be making change to current entry

    my $errmark = "FileQueue::update $id";
    print STDERR $errmark if ($self->{debug});

    # unlock last thing # TODO what if last thing we did was a remove?

    my $fh = $self->{q}->visit($id);
    my $ref = fd_retrieve $fh;

    $ref->{yoke} = 1;

    die "TODO not done yet";
    die "TODO must remove entry from q then store new one";
}

sub showconfig {
    my $self = shift;
    # sanity-check arguments
    for my $c (qw(debug verbose dir )) {
	print "FileQueue arg $c = " . $self->{"$c"} . "\n";
    }
}

sub getparam {
   my ( $self, $name ) = @_;
   print STDERR "FileQueue getparam $name\n" if (defined($self->{debug}));
   my $par = $self->{$name} if ($self->{$name});   
   return $par;
}

1;

=head1 NAME

FileQueue - providing a simple object for safe (lockable) file queue arbritary data

=head1 SYNOPSIS

see SmsEngine::SmsFileQueue, FileQueue.pm source itself, t/filequeue.t

=head2 instantiate and enqueue

 my $q = FileQueue->new( dir => "/tmp/perlFileQueueTest" );
 print $q->{errors}->[0] if ($q->{error});

 my %objhash = ( 'param1' => "complicated", 
 	 	'param2' => "hello foo" );

 my $res = $q->enqueue(\%objhash);

 my $count = $q->show_queue();    # count items in q (from current location)
 my $count = $q->show_queue( 1 ); # print q to screen  (from current location)

 my $count = $q->reset(); # go through full q and reposition us at start

=head2 readnext, dequeue and unlock

 while (1) {

     my $readout = $q->readnext();
     while (defined($readout)) {

         print STDERR "process message from queue\n";
         print STDERR Dumper($readout);

         $res = dosomethingwithitem($readout);

         if ($res =~ "^error" ) { 
             # error => leave item in q
	     $q->unlock(); # unlock last thing (didn't match)
         } else {
             # success => remove item from queue
             my $dqr = $q->dequeue($readout->{id});
	 }

         $readout = $q->readnext(); # get next
     }

     sleep (10);
 }

=head1 CONFIG

 no config file
 pass in dir (directory where q stored)
 optionally debug verbose
 optionally id lock

 see SmsEngine::SmsFileQueue

=head1 DESCRIPTION

 FileQueue combines Queue::Dir and Storable
 Queue::Dir provides interface to queue id hashing and locking.
 Storable provices serialisation of perl structures for storage.

=head1 DESIGN

Interface to lockable dir/file based queue

=head1 COPYRIGHT

Copyright (C) 2004 Doolin Technologies

=cut
