#  SmsEngine.pm
#    - providing a simple object for send/receive sms using
#      various interfaces
#    - load vars from config filesend/receive sms using kannel
#    - this is too thin a layer maybe not worth abstracting out of 
#      SmsEngine.pm
#
#  Copyright (C) 2004 Doolin Technologies
#
#  $Id: SmsEngine.pm,v 1.21 2005/04/27 15:13:26 stella Exp $
#

package SmsEngine;

$VERSION = '0.07';

use strict;
use bytes;
use Config::IniFiles;
use SmsEngine::ReadConfig qw(findConfig);
use SmsEngine::KannelSms;
use SmsEngine::O2ProprietarySms;
use SmsEngine::FakeSms;
use SmsEngine::XiamSms;

sub new($%) {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my %args = @_;

    my $self = {};

    my $errmark = "SmsEngine::new error: ";

    $self->{debug} = 1;
    print STDERR "SmsEngine new\n" if ($self->{debug});

    if ($args{file}) {
	$self->{file} = $args{file};
    } else {
	my $err = $errmark . "Config file arg is mandatory";
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err ]
		 };
    }

    # Read config file (looks at path of binary to find etc dir)
    $self->{config} = &findConfig($self->{file});
    if (!defined($self->{config})) {
	my $err = $errmark . 
	    "cannot find config file $! " . $self->{file};
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err ] };
    }

    $self->{cfg} = new Config::IniFiles( -file => $self->{config} );

    if (!defined($self->{cfg})) {

	# we report errors to STDERR so user can see what is wrong
	my @errors = ( $errmark . "cannot load config file " . $self->{config},
		       @Config::IniFiles::errors );
	
	foreach my $err(@errors) {
	    print STDERR "$err\n";
	}

	# we return errors to higher level scripts so daemons
        # or scripts can take action and warn users in different manner
	return { error => 1, 
		 errors => [ @errors ] };

    }

#TODO: do not die on error, return an error and undef instead

#If an error occurs while parsing the INI file the @Config::IniFiles::errors
#array will contain messages that might help you figure out where the 
#problem is in the file.

    $self->{basedir} = $self->{config};
    $self->{basedir} =~ s/\/[^\/]*$/\//; #filename whack
    $self->{basedir} =~ s/\/[^\/]*\/$/\//; #etc whack

    #$self->{debug} = 1;
    #$self->{verbose} = 1;
    $self->{debug} = ($args{debug})?$args{debug}:0;
    $self->{verbose} = ($args{verbose})?$args{verbose}:0;
    $self->{retry} = $self->{cfg}->val( "commonsms", "retry" );
    $self->{alarmemail} = $self->{cfg}->val( "commonsms", "alarmemail" );
    my $debug = $self->{cfg}->val( "commonsms", "debug" );
    $self->{debug} = $debug if (defined($debug));
    my $verbose = $self->{cfg}->val( "commonsms", "verbose" );
    $self->{verbose} = $verbose if (defined($verbose));
    $self->{io_client_timeout} = $self->{cfg}->val( "commonsms", 
						"io_client_timeout");
    $self->{read_client_timeout} = $self->{cfg}->val( "commonsms", 
						"read_client_timeout");

    $self->{io_client_timeout} = 7 if (!$self->{io_client_timeout});
    $self->{read_client_timeout} = 30 if (!$self->{read_client_timeout});

    $self->{ifacegroup} = $self->{cfg}->val( "commonsms", "ifacegroup" );
    if (!$self->{ifacegroup}) {
	my $err = $errmark . "config group [commonsms] var ifacegroup is mandatory.";
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err,
			     "file: " . $self->{config} ]
		 };
    }
    # note ifacegroup into object overrides default load from commonsms
    #   for testing and non-service-standard script use
    $self->{ifacegroup} = $args{ifacegroup} if ($args{ifacegroup});

    # load config before iface
    if ($self->{ifacegroup}) {
	my $it = loadiface($self,$self->{ifacegroup});
	if ($it->{error}) {

	    my @errors = ( $errmark . "loadiface config error with ifacegroup",
			   @{$it->{errors}},
			   "file: " . $self->{config} );

	    foreach my $err(@errors) {
		print STDERR "$err\n";
	    }

	    return { error => 1,
		     errors => [ @errors ] };
	}
    }

    # load config and iface before account
    if ($self->{accountgroup}) {
	my $at = loadaccount($self,$self->{accountgroup});
	if ($at->{error}) {
	    my @errors = ($errmark . " loadaccount config error with accountgroup",
			  @{$at->{errors}},
			  "file: " . $self->{config} );

	    foreach my $err(@errors) {
		print STDERR "$err\n";
	    }

	    return { error => 1,
		     errors => [ @errors ] };

	}
    }

    #bless _val, $class;
    bless $self, $class;
}

# used b4 self is blessed, remove or use
sub _val {
   my $self = shift;
   my $sname = shift;
   my $vname = shift;

   my $value = $self->{cfg}->val( $sname, $vname );
   if ($self->{verbose}) { print "val $sname-$vname = $value\n"; }
   return $value;
}

# can be called b4 $self is blessed
sub loadiface{
   my $self = shift;
   $self->{iface} = shift;

   if ($self->{verbose}) { print "loadiface ".$self->{iface}."\n"; }

   if (_val( $self, $self->{iface}, "object" ) ne "iface" ) {
       return { error => 1,
		errors => [ "object " . $self->{iface} . " not an interface? " ] };
   }

   my $error = 0;
   my @errors;

   # what account and inout groups to use
   for my $c (qw(accountgroup inoutgroup)) {
       $self->{$c} = _val( $self, $self->{iface}, qq($c) );
       if (!$self->{$c}) {
	   $errors[$error++] = "object $self->{iface} argument $c is mandatory.";
       }
   }

   # then from inout
   if ($self->{inoutgroup}) {
       $self->{inout} = $self->{inoutgroup};
       for my $c (qw(outaddress outport inaddress inport)) {
	   $self->{$c} = _val( $self, $self->{inout}, qq($c) );
	   if (!$self->{$c}) {
	       $errors[$error++] = "inout object $self->{inout} argument $c is mandatory.";
	   }
       }
   }

   if ($error) {
       return { error => 1,
		errors => [ @errors ] };
   }


   # then ifacetype
   $self->{ifacetype} = _val( $self, $self->{iface}, "ifacetype" );

   print STDERR "loadiface ".$self->{iface}.
       "ifacetype is ".$self->{ifacetype}."\n" if ($self->{debug});

   if ($self->{ifacetype} eq 'kannelsms') {
       #TODO this code inside KannelSms.pm
       # $self->{ObjectSms} = KannelSms->new( engine => $self ); 

       for my $c (qw(url user pass)) {
	   $self->{$c} = _val( $self, $self->{iface}, qq($c) );
	   if (!$self->{$c}) {
	       $errors[$error++] = "iface object type $self->{ifacetype} argument $c is mandatory.";
	   }
       }

       if ($error) {
	   return { error => 1,
		    errors => [ @errors ] };
       }

       $self->{kannelsms} = SmsEngine::KannelSms->new( url => $self->{url}, 
						       user => $self->{user}, 
						       pass => $self->{pass});

   } elsif ($self->{ifacetype} eq 'o2proprietary') {
       #TODO this code inside O2ProprietarySms.pm
       $self->{ObjectSms} = SmsEngine::O2ProprietarySms->new( engine => $self,
							      debug => $self->{debug},
							      verbose => $self->{verbose}); 

   } elsif ($self->{ifacetype} eq 'xiamsms') {
       #TODO this code inside XiamSms.pm
       $self->{ObjectSms} = SmsEngine::XiamSms->new( engine => $self,
						     debug => $self->{debug},
						     verbose => $self->{verbose}); 

   } elsif ($self->{ifacetype} eq 'fake') {
       #TODO this code inside FakeSms.pm
       # $self->{ObjectSms} = FakeSms->new( engine => $self ); 

       for my $c (qw(file)) {
	   $self->{$c} = _val( $self, $self->{iface}, qq($c) );
	   if (!$self->{$c}) {
	       $errors[$error++] = "iface object type $self->{ifacetype} argument $c is mandatory.";
	   }
       }

       open(FILE, ">>".$self->{file} ) 
	   or $errors[$error++] = "Can't open ".$self->{file}." for writing $!.";

       if ($error) {
	   return { error => 1,
		    errors => [ @errors ] };
       }

       print FILE "SmsEngine Fake iface instantiated\n";
       close(FILE);

   } else {

       $errors[$error++] = 
	   "i interface " . $self->{iface} . " type " . $self->{ifacetype} . " is not supported";

   }

   if ($error) {
       return { error => 1,
		errors => [ @errors ] };
   } else {
       return { error => 0,
		type => $self->{ifacetype} };
   }
}

# can be called b4 $self is blessed
sub loadaccount{
   my $self = shift;
   $self->{account} = shift;

   if ($self->{verbose}) { print "loadaccount ".$self->{account}."\n"; }

   my $o = _val( $self, $self->{account}, "object" );
   if (!defined($o) || $o ne "account" ) {
       return { error => 1,
		errors => [ "not an account? ".$self->{account} ] };
   }

   $self->{accountifacetype}=_val( $self, $self->{account}, "ifacetype" );
   if ($self->{ifacetype} ne $self->{accountifacetype}) {
       return { error => 1,
		errors => [ "account ifacetype".$self->{accountifacetype}.
			    " doesn\'t match ".$self->{ifacetype} ] };
   }

   if ($self->{ifacetype} eq 'kannelsms') {
       #TODO this code inside KannelSms.pm
       # $self->{ObjectSms}->account( engine => $self ); 
       # kannel gateway server holds kannel account details - not needed here
       #$self->{url} = _val( $self, $self->{account}, "url" );
   } elsif ($self->{ifacetype} eq 'o2proprietary' ||
	    $self->{ifacetype} eq 'xiamsms') {
       #TODO this code inside O2ProprietarySms.pm and XiamSms.pm

       my $act = $self->{ObjectSms}->loadaccount( $self ); 
       if ($act->{error}) {
	   return { error => 1,
		    errors => [ @{$act->{errors}} ] };
       }

       $self->{ObjectSms}->showconfig() if ($self->{verbose});

   } elsif ($self->{ifacetype} eq 'fake') {
       #TODO this code inside FakeSms.pm
       # $self->{ObjectSms}->account( engine => $self ); 
       open(FILE, ">>".$self->{file} ) 
	   or return { error => 1,
		       errors => [ "Can't open ".$self->{file}." for writing." ] };
       print FILE "SmsEngine Fake account loaded\n";
       close(FILE);
   } else {
       return { error => 1,
		errors => [ "account interface ".$self->{account}." type ".$self->{ifacetype}." is not supported" ] };
   }

   return { error => 0,
	    type => $self->{ifacetype} };
}

sub showconfig {

    my $self = shift;
    # sanity-check arguments
    for my $c (qw(debug verbose alarmemail
		  config iface account inout
		  outaddress outport inaddress inport                  
		  )) {
	print "SmsEngine arg $c = " . $self->{"$c"} . "\n";
    }

}

sub getparam {
   my ( $self, $name ) = @_;

   my $par;

   $par = $self->{$name} if ($self->{$name});
   $par = $self->{ObjectSms}->{$name} if ($self->{ObjectSms}->{$name});
   
   return $par;
}

sub sendsms {

   my ( $self, $to, $from, $text ) = @_;
   my %args = @_;

   #TODO OO code
   # this code inside different Sms interface modules
   # my $result = $self->{ObjectSms}->sendsms( engine => $self, 
   #     to => $to, from => $from, text => $text, args => %args ); 
   #  or
   # my $result = $self->{ObjectSms}->sendsms( engine => $self, 
   #     to => $to, from => $from, text => $text, args => %args ); 

   if (!$self->{iface} || !$self->{account}) {
       die "iface ".$self->{iface}." or account ".$self->{account}." not yet set";
   }

   my $result;

   if ($self->{ifacetype} eq 'kannelsms') {
       #TODO this code inside KannelSms.pm
       # $self->{ObjectSms}->account( engine => $self ); 
       # kannel gateway server holds kannel account details - not needed here
       #$self->{url} = $self->_val( $self->{account}, "url" );

       $result = $self->{kannelsms}->sendsms($to,$from,$text);
       print "sendsms result: $result\n";

   } elsif ($self->{ifacetype} eq 'o2proprietary' ||
	    $self->{ifacetype} eq 'xiamsms') {
       #TODO this code inside O2ProprietarySms.pm
       # my $result = $self->{ObjectSms}->account( engine => $self ); 
       print "SmsEngine o2proprietary sendsms to $to from $from text $text\n";

       $result = $self->{ObjectSms}->sendsms($to,$from,$text);
       print "sendsms result: $result\n" if ($self->{verbose});

   } elsif ($self->{ifacetype} eq 'fake') {
       #TODO this code inside FakeSms.pm
       # my $result = $self->{ObjectSms}->account( engine => $self ); 
       open(FILE, ">>".$self->{file} ) 
	   or die "Can't open ".$self->{file}." for writing.";
       print FILE "SmsEngine Fake sendsms to $to from $from text $text\n";
       close(FILE);

   } else {
       die "sendsms interface ".$self->{account}." type ".$self->{ifacetype}." is not supported";
   }

    return $result;
}

1;

=head1 NAME

SmsEngine - providing a simple object for send/receive sms using kannel

=head1 SYNOPSIS

Changed. Changed utterly!

see SmsEngine::O2ProprietarySms

  #!/usr/bin/perl -w

  use strict;
  use SmsEngine::SmsEngine;

  my $sendsms = SmsEngine->new( file => "topupclub_config.ini", 
    iface => "o2proprietary_iface", 
    account => "topupclub_o2_account" );

  my $to = "353861111111";
  my $from = "SmsEngine.pm"; # text from not supported by all smscs/handsets
  my $text = "SmsEngine.pm example text message";
  my $result = $sendsms->sendsms($to,$from,$text);
  print "sendsms result: $result\n";

  undef $sendsms;

  my $sendsms = SmsEngine->new( file => "topupclub_config.ini"); 
  loadiface("kannelsms_iface");
  loadaccount("clickatell_kannel_account" );
  my $result = $sendsms->sendsms($to,$from,$text);
  print "sendsms result: $result\n";

  # not completely safe - could contaminate iface/account BUT ...
  loadiface("fake_iface");
  loadaccount("fake_account" );
  my $result = $sendsms->sendsms($to,$from,$text);
  print "sendsms result: $result\n";

  undef $sendsms;

=head1 CONFIG

[commonsms]
retry = 0
ifacegroup = o2proprietary_iface
# TODO: alarms/alerts if o2 sms servers fail
# other aspects of system fail ? e.g. can't forward to tucd or other filter
alarmemail = jamesc@doolin.com, webmaster@betty.dev.ie.alphyra.com
#debug = 1
#verbose = 1

[o2proprietary_iface]
object = iface
ifacetype = o2proprietary
inoutgroup = smsengine_tucd
accountgroup = topupclub_o2_account_live

[smsengine_tucd]
# server that receives xml from tucd and sends xmlrpc to o2
# sms out localhost, port 8988 from tucd suggested
outaddress = localhost
outport = 8152
# sms in localhost, port 8151 from tucd suggested
inaddress = localhost
inport = 8151

[topupclub_o2_account_live]
object = account
ifacetype = o2proprietary
# The shortcode you have been assigned to use is 353866018440 .
https_soap_url = https://sms-open.o2.co.uk:444/apilistener/SMSOA_soap_listener
https_xmlrpc_url = https://sms-open.o2.co.uk:444/apilistener/SMSOA_xmlrpc_listener
https_xmlrpc_timeout = 15
https_cert_file = /usr/local/smsengine/etc/cacert-live.pem
https_pkcs12_file = /usr/local/smsengine/etc/alphyra_live.p12
https_pkcs12_password = fiqh8523
#https_pkcs12_password = ? get from Umar - ask James/Paul ?
proxy = http://proxy:8080

#NOTE: I suspect test systems within alphyra will all need to use the proxy
#as shown in examples.  Live topupclub system may have different proxy
#OR no proxy (direct connection).

# server that receives xmlrpc from o2 and sends xml to tucd
# localhost, port 8987 from tucd suggested
# server/port settings will be mandated by configuration of 
#   connection from o2 and firewall/routing of traffic from o2 
smsinaddress = localhost
smsinport = 8987

# filter messages to restrict services on free shortcode
smsinfilter = free_filter, non_free_filter, default_filter
# messages from non_free or unknown shortcode (not matching any filter)
# are unrestricted 
# so we can actually do without non_free filter and default_filter
# though if filters are misconfigured default behaviour being to make
# available all services may not be desired

[free_filter]
object=filter
# only allowed to do two actions on free shortcode, otherwise alphyra pays
# shortcode/keywords can be exact strings or could be regular expressions
shortcode = 353863833342
smsinkeywords = REGISTER, VALIDATE
failmessage = "Only REGISTER and VALIDATE services are available on this number. Send TUC text commands to 51101."
# apply this filter first
priority = 1
# if filter catches something it chomps it up and no other filters will catch it
swallow = 1

[non_free_filter]
object=filter
shortcode = 353863832032
smsinkeywords = .*
priority = 2
# neither failmessage or outaddress/outport are set 
# so this filter has no action so it might as well not be here
# is here as an example placeholder 

[default_filter]
object=filter
shortcode = .*
smsinkeywords = .*
#priority = 2

# not used, an example
# TODO: make overriding of sms rx possible
[funny_filter]
object=filter
shortcode = .*
smsinkeywords = funny
# override receiver of sms set in iface group inoutgroup = smsengine_tucd
outaddress = localhost
outport = 8987
swallow = 1

[fake_iface]
object = iface
ifacetype = fake
file = /tmp/fakeSmsEngineIface.log
inoutgroup = smsengine_tucd
accountgroup = fake_account

[fake_account]
object = account
ifacetype = fake

[smskannel_iface]
object = iface
ifacetype = kannelsms
inoutgroup = smsengine_tucd
accountgroup = clickatell_kannel_account
url = http://localhost:13013/cgi-bin/sendsms
user = foo
pass = test

[clickatell_kannel_account]
object = account
ifacetype = kannelsms


=head1 DESCRIPTION

Provide a simple perl module for easy interface to send/receive sms.
Using various interfaces:

  Using kannel sms gateway. http://www.kannel.org
  Obviously kannel software should be installed and running before 
   this part of module is usable.

  Using o2proprietary mms02 MRS XML/RPC interface.
  Scripts to interface to this pre-installed & configured.

  Using fake sendsms for testing.

=head1 DESIGN

Interface to other modules which can send sms.

=head1 COPYRIGHT

Copyright (C) 2004 Doolin Technologies

=cut
