#  SmsFileQueue.pm
#    - using FileQueue.pm object for safe (lockable) file queue arbritary data
#    - validating inserts conform with sms message format for SmsFileQueue.pm
#
#  Copyright (C) 2004 Doolin Technologies
#
#  $Id: SmsFileQueue.pm,v 1.9 2004/11/04 11:22:13 jamesc Exp $
#

package SmsEngine::SmsFileQueue;

use strict;
use bytes;
use FileQueue();
use vars qw(@ISA $VERSION);
@ISA = ("FileQueue");
$VERSION = '0.01';

use Config::IniFiles;
use SmsEngine::ReadConfig qw(findConfig);

# DONE: load config file params
# DONE: tested access to FileQueue methods

# TODO: priorities (long term wish)
# TODO: retry profiles (long term wish)

sub new($%) {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my %args = @_;

    my $errmark = "SmsFileQueue::new error: ";

    my $myself;

    if ($args{file} && $args{queuename}) {
	$myself->{file} = $args{file};
	$myself->{queuename} = $args{queuename};
    } else {
	# TODO: we could behave more nicely 
        # and look for /etc/filequeue.conf or filequeue.conf in current dir
	my $err = $errmark . "Config file arg and queue name are mandatory";
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err ]
		 };
    }

    # Read config file (looks at path of binary to find etc dir)
    $myself->{config} = &findConfig($myself->{file});
    if (!defined($myself->{config})) {
	my $err = $errmark . 
	    "cannot find config file $! " . $myself->{file};
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err ] };
    }

    $myself->{cfg} = new Config::IniFiles( -file => $myself->{config} );

    if (!defined($myself->{cfg})) {

	# we report errors to STDERR so user can see what is wrong
	my @errors = ( $errmark . "cannot load config file " . $myself->{config},
		       @Config::IniFiles::errors );
	
	foreach my $err(@errors) {
	    print STDERR "$err\n";
	}

	# we return errors to higher level scripts so daemons
        # or scripts can take action and warn users in different manner
	return { error => 1, 
		 errors => [ @errors ] };

    }

# do not die on error, return an error and undef instead

#If an error occurs while parsing the INI file the @Config::IniFiles::errors
#array will contain messages that might help you figure out where the 
#problem is in the file.

    $myself->{basedir} = $myself->{config};
    $myself->{basedir} =~ s/\/[^\/]*$/\//; #filename whack
    $myself->{basedir} =~ s/\/[^\/]*\/$/\//; #etc whack

    #$myself->{debug} = 1;
    #$myself->{verbose} = 1;
    $myself->{debug} = ($args{debug})?$args{debug}:0;
    $myself->{verbose} = ($args{verbose})?$args{verbose}:0;

# e.g. 
#[commonsms]
#debug = 0
#verbose = 0
#smsfilequeue = smsfilequeue
#smsfiledonequeue = smsfiledonequeue
#smsengineretryqueue = smsengineretryqueue
#smsenginetotalfailurequeue = smsenginetotalfailurequeue

#[smsfilequeue]
#object = filequeue
#filequeuetype = smsfilequeue
#dir = /usr/local/smsengine/var/spool/smsfilequeue

    my $debug = $myself->{cfg}->val( "commonsms", "debug" );
    $myself->{debug} = $debug if (defined($debug));
    my $verbose = $myself->{cfg}->val( "commonsms", "verbose" );
    $myself->{verbose} = $verbose if (defined($verbose));

    $myself->{queueconfiggroup} = $myself->{cfg}->val( "commonsms", $myself->{queuename} );

    # if that's not there error
    if (!$myself->{queueconfiggroup}) {
	my $err = $errmark . "config group [commonsms] var ".
	    $myself->{queuename}." is mandatory for this script.";
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err,
			     "file: " . $myself->{config} ]
		 };
    }

    $myself->{queuedir} = $myself->{cfg}->val( $myself->{queueconfiggroup}, "dir" );

    # if that's not there error
    if (!$myself->{queuedir}) {
	my $err = $errmark . "config group [".$myself->{queueconfiggroup}.
	    "] var dir is mandatory for this script.";
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err,
			     "file: " . $myself->{config} ]
		 };
    }

    my $self = $class->SUPER::new( dir => $myself->{queuedir} );
    #my $self = FileQueue->new( dir => $myself->{queuedir} );

    $self->{file} = $myself->{file};
    $self->{basedir} = $myself->{basedir};
    $self->{queuedir} = $myself->{queuedir};
    $self->{queuename} = $myself->{queuename};
    $self->{queueconfiggroup} = $myself->{queueconfiggroup};
    # debug verbose ?

    bless $self, $class;
}

# used b4 self is blessed, remove or use
sub _val {
   my $self = shift;
   my $sname = shift;
   my $vname = shift;

   my $value = $self->{cfg}->val( $sname, $vname );
   if ($self->{verbose}) { print "val $sname-$vname = $value\n"; }
   return $value;
}

sub showconfig {
    my $self = shift;

    # call also showconfig of derived-from class
    $self->SUPER::showconfig();

    # sanity-check arguments
    for my $c (qw(debug verbose file queuename
		  queueconfiggroup queuedir
		  )) {
	print "SmsFileQueue arg $c = " . $self->{"$c"} . "\n";
    }
}

sub getparam {
   my ( $self, $name ) = @_;
   #print STDERR "SmsFileQueue getparam $name\n" if (defined($self->{debug}));
   my $par = $self->{"$name"} if (defined($self->{"$name"}));
   $par = $self->SUPER::getparam("$name") if (!defined($par));
   return $par;
}

# inherited methods:

# my $res = $smsfilequeue->enqueue( $req );
# $smsfilequeue->reset();  # so readnext will start with highest priority
# while ($smsfilequeue->notempty()) { # blocks and waits for lock ?
#     my ($id, $req) = $smsfilequeue->readnext();
#	 $smsfilequeue->dequeue($id);
#	 $smsfilequeue->update($id,{ yokeybob => "thingybob" });

1;

=head1 NAME

SmsFileQueue

    read smsengine config file for queue info
    use FileQueue.pm object for safe (lockable) file queue 
     (enqueue or readnext/dequeue)
    particular sms data into queue
    validate enqueues conform with sms message format    

=head1 SYNOPSIS

=head2 Instantiation

All scripts using smsfilequeue will instantiate the object and object
strenuously if cannot do that.

 use SmsEngine::SmsFileQueue;

 my $smsfilequeue = SmsFileQueue->new( 
    file => "smsfilequeue.conf",
    queuename => "smsfilequeue"
    );

 if ($smsfilequeue->{error}) {
     my $errs;
     foreach my $err(@{$smsfilequeue->{errors}}) {
	 print "$err\n";
	 $errs .= "$err\n";
     }
     # TODO how likely is this and tidy up how return stuff
     my $err = "cannot instantiate SmsEngine SmsFileQueue :(".
               " installation or config problem.\n$errs";
     logevent( $MSG_EMERG, $err);
     die $err;
 }

=head2 enqueue

Scripts sending sms using this method fire them into queue and forget about them.
e.g. smsengine/test/smsouttofilequeue.pl

 my $req = { text => "send sms writing to smsengine smsfilequeue", 
             to => [ $testnum ],
	     OtherParam => "miaow" };

 # TODO batch or call other methods, not needed now
 # $req->{batch} = 1;
 # $req->{method} = "querybatchrequest";

 # priority and profile(retry profile) not needed
 my $res = $smsfilequeue->enqueue( $req );

 if ($res->{error} == 0) {
     my $err = "could happen if someone removes the Q directory". 
               " or disk goes away or file system is full.\n".
               "This is TeRRible! Atrocious even! :(";
     logevent( $MSG_EMERG, $err);
 }

Internally in SmsEngine if a send fails the sms goes into the retry queue

 # initial send fails => put sms into the retry queue and forget it
 # fault queue retry daemon takes it from there

 my $res = $smsretryqueue->enqueue( $req );

That's all for scripts for putting things into queue.

=head2 readnext, dequeue and unlock

Scripts processing queues instantiate smsfilequeue object as above.
Use readnext and dequeue.
e.g. smsengine/bin/smsoutfileq.daemon

    while (1) {

        my $readout = $smsfq->readnext();
        while (defined($readout)) {

            print STDERR "process message from queue\n";
            print STDERR Dumper($readout);

            if (defined($readout->{ref}) &&
	        defined($readout->{ref}->{to}) && defined($readout->{ref}->{to}->[0]) &&
  	        defined($readout->{ref}->{text})) {

	        my $message = "<SMS><request><mobile>" . $readout->{ref}->{to}->[0] . "</mobile>".
		    "<msg>" . $readout->{ref}->{text} . "</msg></request></SMS>";	    
 	        my $res = $smsout->send($message);

	        print STDERR "\nsms send $res\n";

	        if ($res =~ "^error" ) { 
                    # error => leave item in q
		    $smsfq->unlock(); # unlock last thing (didn't match)
		    print STDERR "sending error: unlock\n";
   	        } else {
		    my $dqr = $smsfq->dequeue($readout->{id});
  		    print STDERR "success response. dequeue: $dqr\n";
	        }

	    } else {
	        # TODO: add thing to finished/something was wrong queue
	        my $dqr = $smsfq->dequeue($readout->{id});
	        print STDERR "something was wrong $dqr\n";
	}

	$readout = $smsfq->readnext(); # get next

    }

    sleep (10);
    }


=head1 USAGE - SCRIPTS and DAEMONS

SmsFileQueue is used for:

    sms out file queue
    sms fault queue
    sms error queue
    sms done queue

 instantiation may use same config file specifying different queuename:

    my $smsfilequeue = SmsFileQueue->new( 
          file => "smsfilequeue.conf",
          queuename => "smsfilequeue"
          );

    my $smsfilequeue = SmsFileQueue->new( 
          file => "smsfilequeue.conf",
          queuename => "smsfiledonequeue"
          );

For each queue there is at least one script/daemon which may enqueue 
entries. There is another script/daemon which processed thbe queues 
and takes action on each entry.

scripts in smsengine:

    smsengine/test/smsouttofilequeue.pl - enqueue to filequeue
    smsengine/bin/smsoutfileq.daemon    - process filequeue
    smsengine/bin/o2xmlrpcout.daemon    - TODO enqueue to filedonequeue, faultqueue or errorqueue

    TODO: smsengine/bin/o2xmlrpcfaultretry.daemon    - TODO not specified, no requirements
    TODO: smsengine/bin/smserrorprocess.daemon    - TODO not specified, no requirements

A daemon script processes the smsfilequeue.
Other scripts (either daemon or manually run) can process the other queues.
Results of the sms send are also logged into queues so that scripts
calling enqueue on smsfilequeue can later query status of the sms.

The daemon script processing the smsfilequeue sends sms to smsengine.
And then forgets the message.
Retries and fault processing handled in smsengine.
If smsengine is not listening then we do have a problem and Q must hold 
up and wait for service to come back.

Triggered event on directory changes can easily be done with kernel mod but
is much less portable. A simple timeout suffices.

=head1 CONFIG

TND: priorities and retry profiles are long term wishes which possibly are not needed
and would just be a waste of time

In [commonsms] group queue names are specified in form <name> = <queuegroup>.
In this example we keep queue names the same as the name of the group/section
in the config file which defines the queue.

    [commonsms]
    debug = 0
    verbose = 0
    smsfilequeue = smsfilequeue
    smsfiledonequeue = smsfiledonequeue
    smsengineretryqueue = smsengineretryqueue
    smsenginetotalfailurequeue = smsenginetotalfailurequeue

For every queue which is used there must be a queuegroup in the config file 
corresponding the the queuegroup specified in the [commonsms] section.

    [smsfilequeue]
    object = filequeue
    filequeuetype = smsfilequeue
    dir = /usr/local/smsengine/var/spool/smsfilequeue

    [smsfiledonequeue]
    object = filequeue
    filequeuetype = smsfilequeue
    dir = /usr/local/smsengine/var/spool/smsfiledonequeue

TODO: (requirements not defined) for retry queue we may need to configure
various retry options

    [smsengineretryqueue]
    object = filequeue
    filequeuetype = smsprifilequeue
    dir = /usr/local/smsengine/var/spool/smsengineretryqueue
    maxpriority = 5
    profiles = common_prof, immediate_prof, infinate_prof
    # immediate messages must succeed within 22 mins or else fail
    immediate_prof = 20,1,120,2,300,4
    common_prof = 20,1,60,2,240,3,600,6,3600,10
    # fail => try after 20 secs 
    # fail => try after 60 secs 
    # fail => try after 4 min
    # fail => try after 10 min
    # fail => try after 10 min
    # fail => try after 10 min
    # fail => try after 1 hr
    # fail => try after 1 hr
    # fail => try after 1 hr
    # fail => try after 1 hr
    infinate_prof = 20,1,60,2,3600,0

    [smsenginetotalfailurequeue]
    object = filequeue
    filequeuetype = smsfilequeue
    dir = /usr/local/smsengine/var/spool/smsenginetotalfailurequeue


=head1 DESCRIPTION

 perl modules SmsEngine::SmsFileQueue ISA FileQueue 
 FileQueue combines Queue::Dir and Storable
 Queue::Dir provides interface to queue id hashing and locking.
 Storable provices serialisation of perl structures for storage.

When we add a file into Q we must use locking as script waiting for files
is so blindingly fast and efficient it will pick up file as soon as 
is opened for writing. We could make it less blindingly fast and efficient
but backing off and checking for syntatic completeness is just a mess and, 
well, locking is there so we don't have to make a mess like that.
Possibly using "mv" to ensure only entire files are received would work but then again ...

So. Locking.
smsengine has to do the locking itself so.
At least smsengine can also provide a module to do the locking and safe placement
of files into spool queue area.

=head1 DESIGN

Interface to other modules which can send sms.
Perl modules Queue::Dir and Storable are combined in our FileQueue
FileQueue is specialised in SmsEngine::SmsFileQueue

=head1 COPYRIGHT

Copyright (C) 2004 Doolin Technologies

=cut
