#  SmsOutToSmsEngineDaemon.pm
#    - providing a very simple object to send sms using a socket
#    - simple xml out to SmsEngine daemons
#
#  Copyright (C) 2004 Doolin Technologies
#
#  $Id: SmsOutToSmsEngineDaemon.pm,v 1.5 2005/04/27 10:07:12 jamesc Exp $
#

$VERSION = '0.01';

package SmsEngine::SmsOutToSmsEngineDaemon;

use strict;
use bytes;
use Config::IniFiles;
use IO::Socket;
use SmsEngine::ReadConfig qw(findConfig);

sub new($%) {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my %args = @_;

    my $self = {};

    my $errmark = "SmsOutToSmsEngineDaemon new error ";

    if ($args{file}) {
	$self->{file} = $args{file};
    } else {
	my $err = $errmark . "Config file arg is mandatory";
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err ]
		 };
    }

    # Read config file (looks at path of binary to find etc dir)
    $self->{config} = &findConfig($self->{file});
    if (!defined($self->{config})) {
	my $err = $errmark . 
	    "cannot find config file $! " . $self->{file};
	print STDERR "$err\n";
	return { error => 1,
		 errors => [ $err ] };
    }

    $self->{cfg} = new Config::IniFiles( -file => $self->{config} );

    if (!defined($self->{cfg})) {

	# we report errors to STDERR so user can see what is wrong
	my @errors = ( $errmark . "cannot load config file " . $self->{config},
		       @Config::IniFiles::errors );
	
	foreach my $err(@errors) {
	    print STDERR "$err\n";
	}

	# we return errors to higher level scripts so daemons
        # or scripts can take action and warn users in different manner
	return { error => 1, 
		 errors => [ @errors ] };

    }

#TODO: do not die on error, return an error and undef instead

#If an error occurs while parsing the INI file the @Config::IniFiles::errors
#array will contain messages that might help you figure out where the 
#problem is in the file.

    $self->{basedir} = $self->{config};
    $self->{basedir} =~ s/\/[^\/]*$/\//; #filename whack
    $self->{basedir} =~ s/\/[^\/]*\/$/\//; #etc whack

    #$self->{debug} = 1;
    $self->{debug} = ($args{debug})?$args{debug}:0;
    my $debug = $self->{cfg}->val( "commonsms", "debug" );
    $self->{debug} = $debug if (defined($debug));

    $self->{outaddress} = ($args{outaddress})?$args{outaddress}:0;
    $self->{outport} = ($args{outport})?$args{outport}:0;
    my $outaddress = $self->{cfg}->val( "commonsms", "outaddress" );
    $self->{outaddress} = $outaddress if (defined($outaddress));
    my $outport = $self->{cfg}->val( "commonsms", "outport" );
    $self->{outport} = $outport if (defined($outport));

    # default 10 poll interval
    my $outpolltime = $self->{cfg}->val( "commonsms", "outpolltime" );
    $outpolltime = 10 if (!defined($outpolltime));
    $self->{outpolltime} = $outpolltime;
    $self->{sms_out_timeout} = $self->{cfg}->val( "commonsms",
                                                    "sms_out_timeout");
    $self->{sms_out_timeout} = 6 if (!$self->{sms_out_timeout});

    if (!$self->{outport} || !$self->{outaddress}) {
	$self->{ifacegroup} = $self->{cfg}->val( "commonsms", "ifacegroup" );
	if (!$self->{ifacegroup}) {
	    my $err = $errmark . "config group [commonsms] var ifacegroup is mandatory.";
	    print STDERR "$err\n";
	    return { error => 1,
		     errors => [ $err,
				 "file: " . $self->{config} ]
				 };
	}
	# note ifacegroup into object overrides default load from commonsms
	#   for testing and non-service-standard script use
	$self->{ifacegroup} = $args{ifacegroup} if ($args{ifacegroup});
	
	# load config before iface
	if ($self->{ifacegroup}) {
	    my $it = loadiface($self,$self->{ifacegroup});
	    if ($it->{error}) {

		my @errors = ( $errmark . "loadiface config error with ifacegroup",
			       @{$it->{errors}},
			       "file: " . $self->{config} );

		foreach my $err(@errors) {
		    print STDERR "$err\n";
		}

		return { error => 1,
			 errors => [ @errors ] };
	    }
	}
    }

    #bless _val, $class;
    bless $self, $class;
}

# used b4 self is blessed, remove or use
sub _val {
   my $self = shift;
   my $sname = shift;
   my $vname = shift;

   my $value = $self->{cfg}->val( $sname, $vname );
   if ($self->{debug}) { print STDERR "val $sname-$vname = $value\n"; }
   return $value;
}

# can be called b4 $self is blessed
sub loadiface{
   my $self = shift;
   $self->{iface} = shift;

   if ($self->{debug}) { print STDERR "loadiface ".$self->{iface}."\n"; }

   if (_val( $self, $self->{iface}, "object" ) ne "iface" ) {
       return { error => 1,
		errors => [ "object " . $self->{iface} . " not an interface? " ] };
   }

   my $error = 0;
   my @errors;

   # what inout group to use
   for my $c (qw(inoutgroup)) {
       $self->{$c} = _val( $self, $self->{iface}, qq($c) );
       if (!$self->{$c}) {
	   $errors[$error++] = "object $self->{iface} argument $c is mandatory.";
       }
   }

   # then from inout
   if ($self->{inoutgroup}) {
       $self->{inout} = $self->{inoutgroup};
       for my $c (qw(outaddress outport)) {
	   $self->{$c} = _val( $self, $self->{inout}, qq($c) );
	   if (!$self->{$c}) {
	       $errors[$error++] = "inout object $self->{inout} argument $c is mandatory.";
	   }
       }
   }

   if ($error) {
       return { error => 1,
		errors => [ @errors ] };
   } else {
       return { error => 0,
		type => $self->{ifacetype} };
   }
}

sub showconfig {

    my $self = shift;
    # sanity-check arguments
    for my $c (qw(debug config iface inout
		  outaddress outport outpolltime sms_out_timeout
		  )) {
	print "SmsOutToSmsEngineDaemon arg $c = " . $self->{"$c"} . "\n";
    }

}

sub getparam {
   my ( $self, $name ) = @_;

   my $par;

   $par = $self->{$name} if ($self->{$name});
   $par = $self->{ObjectSms}->{$name} if ($self->{ObjectSms}->{$name});
   
   return $par;
}

sub send {

   my ( $self, $msg ) = @_;
   my %args = @_;

   print STDERR "create socket " . $self->{outaddress} . ":" . 
       $self->{outport} . "\n" if ($self->{debug});
   my $sock = new IO::Socket::INET (PeerAddr => $self->{outaddress},
				 PeerPort => $self->{outport},
				 Proto    => 'tcp'
				 );
   if (!defined($sock)) {
       my $err = "error Socket could not be created. Reason $!\n";
       print STDERR "$err\n";
       return $err;
   }

   # should always be defined
   if(defined($msg)) {
       print STDERR "\nwriting \"$msg\" to socket\n" if ($self->{debug});
       print $sock $msg;
   }

   my $timeout = $self->{sms_out_timeout};
   $timeout = 6 if (!defined($timeout));
   
   my $start = time();
   my $line;
   my $now = $start;
   my $result = "";

   while($now-$start<$timeout){
       $line = <$sock>;
       if ($line) {
	   print "read: \"".$line."\"\n" if ($self->{debug});
	   $result .= $line;
       }
       $now=time();
   }
   
   close ($sock);

   return $result;
}

1;

=head1 NAME

SmsOutToSmsEngineDaemon
    - providing a very simple object to send sms using a socket
    - simple xml out to SmsEngine daemons

=head1 SYNOPSIS

use SmsEngine::SmsOutToSmsEngineDaemon;
my $smsout = SmsOutToSmsEngineDaemon->new( file => "smsengine.topupclub.conf");

my $invalidmobilenumber = "hi mom";
my $message = "<SMS><request><mobile>$invalidmobilenumber</mobile>".
	"<msg>TUC smsengine test-o2xmlrpcout.daemon invalid mobile number</msg></request></SMS>";
my $res = $smsout->send($message);

my $tn = "check smsout response";
::is( $res !~ '^error', 1, $tn);
::is( $res =~ "<SMS><response><errorcode>0</errorcode><msg>Success</msg>", 1, $tn);
::is( $res =~ "<mobileerror><mobile>$invalidmobilenumber</mobile>", 1, $tn);
::is( $res =~ "<faultcode>202</faultcode>", 1, $tn);
::is( $res =~ "</mobileerror></response></SMS>", 1, $tn);

=head1 CONFIG

[commonsms]
retry = 0
accountgroup = topupclub_o2_account
ifacegroup = o2proprietary_iface

[o2proprietary_iface]
object = iface
ifacetype = o2proprietary
inoutgroup = smsengine_tucd
accountgroup = topupclub_o2_account_live

[smsengine_tucd]
# server that receives xml from tucd and sends xmlrpc to o2
# sms out localhost, port 8988 from tucd suggested
outaddress = localhost
outport = 8152
# sms in localhost, port 8151 from tucd suggested
inaddress = localhost
inport = 8151

=head1 DESCRIPTION

Provide a simple perl module for easy interface to send sms to smsengine daemon port.

=head1 COPYRIGHT

Copyright (C) 2004 Doolin Technologies

=cut
