#!/usr/bin/perl -w # system calls, readdir, unlink have problems with TAINT
#!/usr/bin/perl -Tw # system calls have problems with TAINT

# test SmsFileQueue with multiple processes concurrently writing and reading into queues
# tests that SmsFileQueue is really multiple process safe
# doesn't have to be thread safe
#  but might accidentally test that too

# make this an optional test ?
# use POE ?
# fork off a couple of children and use them

use strict;
use bytes;

use lib qw(t/lib);
use Test::More tests =>
    1 +          # use_ok
    6 +          # SmsFileQueue create and check
    2 +          # write to queues (children will write too but different process)
    8 +          # read all queue contents and check everything is there
    0 +          # TODO writes while reading
    0 +          # TODO writes while reading and dequeuing
    0 +          # concurrent reading and dequeuing
    1            # read all queue contents and check everything is removed
    ;

BEGIN { use_ok('SmsEngine::SmsFileQueue') }

my $ti = 1;
my $tn = "";
sub registerTest {
    $ti++;
    $tn = shift;
}


# create dir and check is ready for us to write to it

# nuke old test data
if ( -d "/tmp/smsenginefilequeuetest") {
    my $rmtestdircom = "rm -rf /tmp/smsenginefilequeuetest";
    system ($rmtestdircom) == 0 || die "cannot $rmtestdircom $!";
}
::fail "rmdir test setup" if ( -d "/tmp/smsenginefilequeuetest");

# create empty queue dir
my $testdir = "/tmp/smsenginefilequeuetest/var/spool/smsfilequeue";
my $mkdircom = "mkdir -p $testdir";
system ($mkdircom) == 0 || die "cannot $mkdircom $!";
if ( ! -d $testdir) {
    die "cannot create or access $testdir ?$!?";
}



#### okay

# instantiate okay
registerTest("SmsFileQueue create okay");
my $smsfq = SmsEngine::SmsFileQueue->new( debug => 1, verbose => 1,
					  file => "smsfilequeue.conf",
					  queuename => "smsfilequeue");
::ok( defined($smsfq), $tn );
::is( $smsfq->{error}, undef, $tn);
::is( $smsfq->{errors}->[0], undef, $tn);
::is( $smsfq->getparam("file"), "smsfilequeue.conf", $tn );
::is( $smsfq->getparam("dir"), 
      '/tmp/smsenginefilequeuetest/var/spool/smsfilequeue', $tn );
::is( $smsfq->getparam("queuedir"), 
      '/tmp/smsenginefilequeuetest/var/spool/smsfilequeue', $tn );

if ($smsfq->{error}) {
    my $errs;
    foreach my $err(@{$smsfq->{errors}}) {
	print "$err\n";
	$errs .= "$err\n";
    }
    my $err = "cannot instantiate SmsFileQueue :( installation or config problem.\n$errs";
    die $err;
}



# I've an experiment called forkit.pl
# started from lockarea from perl cookbook

my $FORKS = 7;
my $INSERTS = 2;
my $SLEEP;
my $ROO = 42;

my $progenitor = $$;

# install signal handler for children
BEGIN {
    my $called = 0;

    sub genocide {
	exit if $called++;
	print "$$: Time to die, kiddies.\n" if $$ == $progenitor;
	my $job = getpgrp();
	$SIG{INT} = 'IGNORE';
	kill -2, $job if $job;  # killpg(SIGINT, job)
	1 while wait > 0;
	print "$$: My turn\n" if $$ == $progenitor;
	exit;
    }

}

# END would clobber Test::More's test summary and continuing tests
#END { &genocide }

my $ch = 0;
my @children;
my $forks = $FORKS;
while ( $forks-- > 0 && $progenitor == $$ ) {
    fork;
    $children[$ch++] = $forks;
}
# ONLY the parent forks
$forks++; # $forks from 0 to $FORKS - 1
print "hello from $$, my forks is $forks". 
    ( ($progenitor == $$)?" ** I AM THE PROGENITOR ** ":"" )."\n";

if ($progenitor == $$) {
    $SIG{INT} = \&genocide;
} else {
    $SIG{INT} = sub { die "goodbye from $$" };
}

for (my $i=0;$i<$INSERTS;$i++) {

    my $l = int rand($ROO);

    my $msg = "test message $$ $forks $i $l";
    my $to = "badtonumber";

    my $req = { text => qq($msg),
		to => [ $to ],
                pid => $$,
                forks => $forks,
                i => $i };

    my $res = $smsfq->enqueue( $req );

    # test check in here (multiple concurrent processes) confuses Test::More
    if ($progenitor == $$) {
	::is( $res->{error}, 0, $tn);
    }

    if ($res->{error} != 0) {
	my $err = "could happen if someone removes the Q directory". 
	    " or disk goes away or file system is full.\n".
	    $res->{errors}->[0];
	warn $err;
    }

    sleep $SLEEP if $SLEEP;
}

# exit all children

if ($progenitor == $$) {
} else {
    exit 0;
}

### verify items are enqueued

# if we don't wait locks are still in place so we do not read everything
# wait for ALL children
{} while (wait>0);
#$forks = $FORKS;
#while ( $forks-- > 0) {
#    my $wr = wait; # $ch--;
#    print STDERR "wait $wr\n";
#}
#sleep 2;

sub readallQ {
    my $q = shift;
    my $action = shift;

    $q->reset();
 
    my $i = 0;
    my @readout;
    $readout[$i++] = $q->readnext();
    while( defined($readout[$i-1])) {
        $q->unlock(); # unlock last thing (didn't match)
        $readout[$i++] = $q->readnext();
    }

    $q->unlock(); # unlock last thing
    # DON't unlock last thing yet if want to do something with it

    return \@readout;
}

my $readout = readallQ( $smsfq );

#use Data::Dumper;
#print STDERR Dumper($readout);

my $total=0;
my @totpid;
for my $item (@$readout) {
    if (defined($item)) {
	$total++;
	$totpid[$item->{ref}->{forks}]++;
    }
}

::is($total,$INSERTS*($FORKS+1),"all processes inserted $INSERTS items");
::is($totpid[0],2 * $INSERTS,"process fork 0 (parent and one child) inserted $INSERTS items");
for my $f (1 .. $FORKS-1) {
    ::is($totpid[$f],$INSERTS,"process fork $f inserted $INSERTS items");
}


### each child removes entries it made
# TODO: dequeue items, multiple concurrent processes again

sub findinQ {
    my $q = shift;
    my $key = shift;
    my $value = shift;

    $q->reset();
 
    my $i = 0;
    my $readout = $q->readnext();
    $i++;
    while( defined($readout) && $readout->{ref}->{$key} ne $value) {
        $q->unlock(); # unlock last thing (didn't match)
        $readout = $q->readnext();
	$i++;
    }

    #$q->unlock(); # unlock last thing
    # DON't unlock last thing yet if want to do something with it

    return $readout;
}

$forks = $FORKS;
while ( $forks-- > 0 && $progenitor == $$ ) {
    fork;
}

my $readout0 = findinQ( $smsfq, "forks", $forks );
$smsfq->dequeue($readout0->{id});
print STDERR "dequeue $forks\n";
my $readout1 = findinQ( $smsfq, "forks", $forks );
$smsfq->dequeue($readout1->{id});
print STDERR "dequeue $forks\n";

# exit all children

if ($progenitor == $$) {
} else {
    exit 0;
}

# wait or else things will be still undeleted/locked
# wait for ALL children
{} while (wait>0);

# Final wrap up

$readout = readallQ( $smsfq );
$total=0;
for my $item (@$readout) {
    if (defined($item)) {
	$total++;
    }
}

TODO: {
    local $TODO = "test code not fork safe";
    ::is($total,0,"everything removed from q");
}

# darn. dequeue especially not fork safe
# cross-process safe yes. but not in script that forks but shares stuff like this
